/* File:      parm.h
   Date:      3/27/01 (revised from fall 2001 version)
   Author:    C. Paciorek
   Purpose:   header file for parm.C
*/


typedef double(*transform_fxn_ptr) (double);    /* used as type when passing transform fxns */

typedef double(*prior_dens_fxn_ptr) (double,double,double);    /* used as type when passing log density functions; first arg is value to evaluate at, 2nd is first parm, 3rd is second parm */ 

class chain;                          /* chain is used as friend */
class dataset;
class sparse_cov_mat_nonstat;
class sparse_cov_mat_stat;
class hyperparm_vec;
class hyperparm_scal;
class covariate_info;

extern int prior;
extern covariate_info* covariates;
extern dataset* data;
extern int* idum;        /* seed for random number generation during proposals */

/* values used to change proposal sd for different parts of scalar parameter space */
const double highkappa_scale=2.7;/* scale proposal sd by this for high values of kappa */
const double kappa_highval=1.0;  /* use highkappa_scale above this value */
const double lowkappa_scale=0.5; /* scale proposal sd by this for low values of kappa */
const double kappa_lowval=0.4;  /* use lowkappa_scale below this value */

const double nu_lowval=1.5;     /*1.5 use lownu_scale below this value */
const double lownu_scale=0.1;   /* scale proposal sd by this for low values of nu */
const double nu_highval=7.0;    /* 7.0 use highnu_scale above this value */
const double highnu_scale=4.0;  /* 4.0 scale proposal sd by this for high values of nu */

/* Parameters are coded as having a 2-level hierarchy, with the higher level 
being hyperparm_scal and hyperparm_vec, the basic classes scalar and vector classes and the lower level being spatial_parm and nonspatial_parm.  spatial_parm is a vector of parameters that are spatially correlated (i.e. correlated based on the closeness of the covariates), with hyperparameters tying them together. */

/* variables named omega  are on the scale of the proposals and are usually proposed using normal distributions; those named psi are the transformed values of omega used in the likelihood; those named ksi are matrix multiplications of omega that are not yet transformed and/or have not yet had a mean added; for hyperparm_scal, base_value is on scale of proposal and value is after transformation */

/* individual scalar parameters */
class nonspatial_parm{            
public:
  nonspatial_parm(int which_in,double omega_init,double omega_mean, double omega_sd,double prop_sd_in,transform_fxn_ptr transform_in);
  ~nonspatial_parm(void);
  int step(int fixed);                         /* one MCMC proposal */
  int accept(double logprior_diff);            /* decide whether to accept proposal */
  friend chain;                                 /* needs access to psi to pass to data (for noise sd) */
private:
  hyperparm_scal* psi;                        
  int which;                                    /* constant indicating which parameter it is */
  int num_accept;        /* total number accepted */
  int run_num_accept;    /* number accepted in blocks of 100 */
};

/* spatially correlated vector of parameters */
class spatial_parm{                    
public:
  spatial_parm(int which_in, double* omega_init,double mu_init, double logsigma_init,double nu_init,double mu_mean, double mu_sd, double logsigma_mean, double logsigma_sd, double nu_mean,double nu_sd,double omega_prop_sd, double mu_prop_sd, double logsigma_prop_sd, double nu_propsd, transform_fxn_ptr transform_in);
  spatial_parm(int which_in, double* logkappa_init, double* gamma_init,double nu_init, double logkappa_parm1,double logkappa_parm2,double nu_mean,double nu_sd,double* logkappa_prop_sd,double* gamma_prop_sd, double nu_propsd);
  ~spatial_parm(void);
  void step(int mode,int langevin,int kappa_fixed,int gamma_fixed,int mu_fixed,int sigma_fixed,int nu_fixed,int omega_fixed);      /* one MCMC proposal */
  void ksi_to_psi(void);        /* calculates transformation to original scale */
  double ksi_propose(int update_chi);  /* proposes ksi using posterior mean centering */
  void omega_to_ksi(void);             /* calculate ksi (vector of correlated process values on transformed scale) from a priori independent components omega */
  double change(int caller,int mode);         /* called from code dealing with e eigenstructure  to change fun values */
  void update(int accepted,int mode);         /* called from code dealing wit heigenstructure to update fun */

  /* add information that could not be included in construction */
  void add_child(sparse_cov_mat_stat* cov_in,sparse_cov_mat_stat* cov_full_in);
  void add_child(sparse_cov_mat_nonstat* cov_in,sparse_cov_mat_nonstat* cov_full_in);
  void add_child(spatial_parm* fun_in);
  void add_child(spatial_parm** eigenval_in,spatial_parm** eigenvec_in);


  int accept_nolik(double logprior_diff);               /* determines whether proposal is accepted for proposals not involving likelihood  */
  int accept_lik(double logprior_diff);               /* determines whether proposal is accepted for proposals involving likelihood */
  void sample_full(void);                       /* sample process values at training and test points */
  friend chain;                                 /* needs access to psi to pass value to data */
private:
  int which;                                    /* constant indicating which parameter it is */
  double lam_max;                               /* largest value of eigenvalue allowed; this is done to prevent wandering in part of space where eigenvalues are very large and changing them has little influence on likelihood */
  int num_omegapoints;                          /* num_datapoints otherwise */
  spatial_parm* fun;                            /* used by regr fxn covariance parameters; this is a pointer to the regr fxn object */
  spatial_parm** eigenval;                      /* pointer to eigenvalues used by kernel to change eigen parms */ 
  spatial_parm** eigenvec;                      /* pointer to eigenvector coordinates used by kernel to change eigen parms */ 

  hyperparm_scal** kappa;                       /* sd of kernels (sqrt of eigenvals in mother kernel) */
  hyperparm_scal** gamma;                       /* Givens angles in mother kernel */
  hyperparm_vec* omega;                         /* transformed independent components:  w~N(0,1) prior values */
  hyperparm_vec* psi;                           /* process on original scale */
  double* psi_full;                             /* sample of process on training and test sets */
  hyperparm_vec* ksi;                           /* process on transformed scale  */
  hyperparm_vec* chi;                           /* de-correlated residual about the conditional posterior mean, used in PMC approach */
  hyperparm_vec* chinext;                       /* chi values at proposed ksi under PMC approach */
  double** kappa_dbl;                           /* vector of pointers to kappa->value used as argument for kernel or cov */
  double** gamma_dbl;                           /* vector of pointers to gamma->value used as argument for kernel or cov */
  double* gamma_max;                            /* upper limit of Givens angles, PI */
  double* gamma_min;                            /* lower limit of Givens angles, -PI */

  /* cov matrices for training set */
  sparse_cov_mat_stat* stat_cov;
  sparse_cov_mat_nonstat* nonstat_cov;
  /* cov matrices for training plus test set; leading block is the same as the matrix for the training set */
  sparse_cov_mat_stat* stat_cov_full;
  sparse_cov_mat_nonstat* nonstat_cov_full;

  hyperparm_scal* sigma;                        /* log(sd) of ksi, the transformed process */
  hyperparm_scal* mu;                           /* mean of ksi, the transformed process */
  hyperparm_scal* nu;                           /* smoothness parameter of Matern correlation of process */

  transform_fxn_ptr transform;                  /* transformation function for the parameter */

  /* objects to keep track of number accepted */
  int* num_accept_kappa;
  int* num_accept_gamma;
  int num_accept_omega;
  int* run_num_accept_kappa;
  int* run_num_accept_gamma;
  int run_num_accept_omega;
  int num_accept_mu;
  int num_accept_sigma;
  int run_num_accept_mu;
  int run_num_accept_sigma;
  int num_accept_nu;
  int run_num_accept_nu;

  double* tmpvec;
  double* tmpvec2;
  double* tmpvec_test;
  double* tmpvec_test2;

  /* gradient in Langevin calculations */
  double* newgrad;
  double* oldgrad;
  /* number of times max gradient exceeded */
  int oldgrad_exceeded;
  int newgrad_exceeded;

  /* used in PMC calculations */
  double* LCsum;
  double* Csum_diff;

  double logprior_diff;

 };
   
/* hyperparameter vector with independent components */
/* this is used as a component of the spatial_parm class */
class hyperparm_vec{                         
public:
  hyperparm_vec(int size_in,double* omega_init,double prop_sd_in); /* constructor used for omega parameter vector */
  hyperparm_vec(int size_in);       /* use for psi and ksi, i.e. for parameters that don't need to be proposed and whose initial values are based on omega */
  ~hyperparm_vec(void);
  void update(int accepted);                     /* update current and stored parameter values based on whether proposal accepted */
  void propose(double alternate_prop_sd);        /* generate proposal value */
  double evaluate_prior(void);                   /* calc difference of log prior density of proposal value compared to current value */
  friend chain;                                  /* needs access to logprior */
  friend spatial_parm;                           /* needs access to prop_sd */
private:
  int size;                                      /* number of elements in vector */
  transform_fxn_ptr transform;                   /* g transformation function */
  double prop_sd;                                /* proposal sd */
  double logprior;                               /* current/proposed log prior density */
  double logprior_stored;                        /* density of most recently accepted value */
  double* value;                                 /* current/proposed value */
  double* value_stored;                          /* most recently accepted value */
};

/* scalar hyperparameter -- used as component of spatial_parm and nonspatial_parm */
class hyperparm_scal{                            
public:
  hyperparm_scal(double init,double parm1_in,double parm2_in,double prop_sd_in,prior_dens_fxn_ptr prior_dens_fxn_in,transform_fxn_ptr transform_in);
  void update(int accepted);
  double evaluate_prior(void);
  void propose(double alternate_prop_sd);
  friend spatial_parm;                          /* access to value and value_stored */
  friend chain;                                 /* access to value and value_stored */       
private:
  prior_dens_fxn_ptr prior_dens_fxn;            /* form of prior density */
  transform_fxn_ptr transform;                  /* g transformation function */
  int fixed;                                    /* is parameter fixed or should it be sampled */
  double prop_sd;
  double logprior;
  double logprior_stored;
  double base_value;                            /* value on transformed scale */
  double base_value_stored; 
  double value;                                 /* value on original scale */
  double value_stored;                          
  double parm1;                                 /* prior mean or lower limit of parameter on transformed scale  */
  double parm2;                                 /* prior sd or upper limit of parameter on transformed scale */
};













